/*
 * Copyright (c) 2007 Alexey Vatchenko <av@bsdua.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <sys/types.h>
#include <sys/stat.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

static char *headername = NULL;
static char *structname = NULL;
static char *inputfile = NULL;
static char *funcprefix = "fsss";
static int icase = 0;	/* case sensetive by default */

__dead static void usage(void);

/*
 * Line storing framework
 */
struct newline {
	char **key;	/* allocated unescaped string */
	size_t n_key;
	char *val;	/* just a pointer in task file buffer */
	size_t vallen;
};

static char *	_next_quote(const char *buf, const char *lim);
static char *	_next_brace(const char *buf, const char *lim);
static char *	_next_lf(const char *buf, const char *lim);

void	newline_init(struct newline *nl);
void	newline_final(struct newline *nl);
void	newline_drain(struct newline *nl);
int	newline_addkey(struct newline *nl, const char *start, size_t len);
void	newline_addval(struct newline *nl, const char *start, size_t len);
int	newline_parse(struct newline *nl, const char *buf, const char *lim,
	    char **last);

/*
 * Sorting framework
 */
struct node {
	char *key;
	char *val;
	size_t vallen;
	struct node *orig;	/* If this node is alias, ``orig'' points to
				 * the original node */
	size_t index;		/* index of value (for ``orig'' keys) */
};

struct node_array {
	struct node **node;
	size_t n_node;
};

static int	nodecmp(const void *n1, const void *n2);

void	nodes_init(struct node_array *na);
void	nodes_final(struct node_array *na);
int	nodes_append(struct node_array *na, struct newline *nl);
int	nodes_parse(struct node_array *na, const char *buf, size_t size);
void	nodes_sort(struct node_array *na);
int	nodes_save(struct node_array *na, FILE *cfile, FILE *hfile);

static int	validate_code(const char *str);

static const char hfile_fmt[] =
	"/*\n"
	" * This file is generated by fsss program.\n"
	" * If you want to modify this file modify \"%s\" instead.\n"
	" */\n"
	"#ifndef _FSSS_%s_H_\n"
	"#define _FSSS_%s_H_\n"
	"#include <sys/types.h>\n"
	"#include \"%s\"\n"
	"\n"
	"__BEGIN_DECLS\n"
	"struct %s * %s_%s_search(const char *key);\n"
	"__END_DECLS\n"
	"\n"
	"#endif /* !_FSSS_%s_H_ */\n";

static const char cfile_fmt1[] = 
	"/*\n"
	" * This file is generated by fsss program.\n"
	" * If you want to modify this file modify \"%s\" instead.\n"
	" */\n"
	"#include <sys/types.h>\n"
	"#include <string.h>\n"
	"#include \"%s\"\n"
	"\n"
	"struct fsss_%s_key {\n"
	"	char *key;\n"
	"	struct %s *data;\n"
	"};\n"
	"\n"
	"static struct %s fsss_%s_array[] = {\n";

static const char cfile_fmt2[] =
	"};\n"
	"\n"
	"static struct fsss_%s_key fsss_%s_key_array[] = {\n";

static const char cfile_fmt2dot5[] =
	"\t{\"%s\", fsss_%s_array + %lu},\n";

static const char cfile_fmt3[] =
	"};\n"
	"\n"
	"struct %s * %s_%s_search(const char *key);\n"
	"\n"
	"struct %s *\n"
	"%s_%s_search(const char *key)\n"
	"{\n"
	"	size_t start, end, pos;\n"
	"	int res;\n"
	"\n"
	"	start = 0;\n"
	"	end = sizeof(fsss_%s_key_array) / sizeof(*fsss_%s_key_array) - 1;\n"
	"\n"
	"	for (;;) {\n"
	"		pos = (end - start) / 2 + start;\n"
	"		res = str%scmp(key, fsss_%s_key_array[pos].key);\n"
	"		if (res == 0)\n"
	"			return (fsss_%s_key_array[pos].data);\n"
	"		else if (res > 0) {\n"
	"			if (pos == end)\n"
	"				return (NULL);\n"
	"			start = pos + 1;\n"
	"		} else {\n"
	"			if (pos == start)\n"
	"				return (NULL);\n"
	"			end = pos - 1;\n"
	"		}\n"
	"	}\n"
	"	/* NOTREACHED */\n"
	"}\n";

size_t	read_task_file(const char *filename, char **buf);

__dead static
void usage(void)
{
	extern char *__progname;

	fprintf(stderr,
	    "usage: %s [-c] [-p prefix] struct_name header_file input_file\n",
	    __progname);
	exit(2);
}

static char *
_next_quote(const char *buf, const char *lim)
{
	char *p;
	int eflag;

	p = (char *)buf;
	eflag = 0;
	while (p < lim) {
		if (*p == '\\') {
			eflag = !eflag;
			if (eflag) {
				p++;
				continue;
			}
		}
		if (!eflag && *p == '\"')
			return (p);
		p++;
		eflag = 0;
	}
	return (NULL);
}

static char *
_next_brace(const char *buf, const char *lim)
{
	char *p;
	size_t n;

	n = 0;
	p = (char *)buf;
	while (p < lim) {
		if (*p == '}') {
			if (n == 0)
				return (p);
			n--;
		} else if (*p == '{')
			n++;
		p++;
	}
	return (NULL);
}

static char *
_next_lf(const char *buf, const char *lim)
{
	char *p;

	p = (char *)buf;
	while (p < lim) {
		if (*p == '\n')
			return (p);
		p++;
	}
	return (NULL);
}

void
newline_init(struct newline *nl)
{
	memset(nl, 0, sizeof(*nl));
}

void
newline_final(struct newline *nl)
{
	size_t i;

	if (nl->n_key > 0) {
		for (i = 0; i < nl->n_key; i++)
			free(nl->key[i]);
		free(nl->key);
	}
}

void
newline_drain(struct newline *nl)
{
	if (nl->n_key > 0) {
		free(nl->key);
		nl->key = NULL;
		nl->n_key = 0;
	}
	nl->val = NULL;
	nl->vallen = 0;
}

int
newline_addkey(struct newline *nl, const char *start, size_t len)
{
	char **oldkey, *p;
	size_t i;
	int eflag;

	oldkey = nl->key;
	nl->key = realloc(nl->key, (nl->n_key + 1) * sizeof(*nl->key));
	if (nl->key == NULL) {
		nl->key = oldkey;
		warnx("can't allocate memory");
		return (ENOMEM);
	}

	/*
	 * Allocate size of escaped string. In the worst case we will have
	 * double real length.
	 */
	nl->key[nl->n_key] = (char *)malloc(len + 1);
	if (nl->key[nl->n_key] == NULL) {
		warnx("can't allocate memory");
		return (ENOMEM);
	}

	/*
	 * Unescape string.
	 * \ + \r or \ + \n means nothing. They are used only for indent.
 	 */
	eflag = 0;
	p = nl->key[nl->n_key];
	for (i = 0; i < len; i++) {
		if (start[i] == '\\') {
			eflag = !eflag;
			if (eflag)
				continue;
		}
		if (eflag) {
			if (start[i] != '\r' && start[i] != '\n')
				*p++ = start[i];
		} else
			*p++ = start[i];
		eflag = 0;
	}
	*p = '\0';
	nl->n_key++;
	return (0);
}

void
newline_addval(struct newline *nl, const char *start, size_t len)
{
	nl->vallen = len;
	nl->val = (len > 0) ? (char *)start : NULL;
	if (nl->val == NULL)
		warnx("empty value");
}

int
newline_parse(struct newline *nl, const char *buf, const char *lim, char **last)
{
	char *p, *nxt;
	int error;

	p = (char *)buf;

	for (;;) {
		/* Skip leading spaces */
		while (p < lim && isspace(*p))
			p++;
		if (p >= lim) {
			*last = p;
			if (nl->n_key > 0)
				warnx("ignored keys without values");
			return (0);
		}

		if (*p == '\"') {
			p++;
			if (p >= lim || (nxt = _next_quote(p, lim)) == NULL) {
				warnx("can't find closing quote");
				return (EINVAL);
			}
			error = newline_addkey(nl, p, nxt - p);
			if (error != 0)
				return (error);
			p = nxt + 1;
		} else if (*p == '{') {
			if (nl->n_key == 0) {
				warnx("value without key");
				return (EINVAL);
			}

			p++;
			if (p >= lim || (nxt = _next_brace(p, lim)) == NULL) {
				warnx("can't find closing brace");
				return (EINVAL);
			}
			newline_addval(nl, p, nxt - p);
			*last = nxt + 1;
			return (0);
		} else if (*p == '#') {
			/* Comments */
			nxt = _next_lf(p, lim);
			if (nxt == NULL)
				p = (char *)lim;
			else
				p = nxt + 1;
		} else {
			warnx("illegal symbol");
			return (EINVAL);
		}
	}
	/* NOTREACHED */
}

static int
nodecmp(const void *v1, const void *v2)
{
	struct node *n1, *n2;

	n1 = *(struct node **)v1;
	n2 = *(struct node **)v2;

	return (icase) ? strcasecmp(n1->key, n2->key) :
	    strcmp(n1->key, n2->key);
}

void
nodes_init(struct node_array *na)
{
	na->node = NULL;
	na->n_node = 0;
}

void
nodes_final(struct node_array *na)
{
	size_t i;

	for (i = 0; i < na->n_node; i++) {
		free(na->node[i]->key);
		free(na->node[i]);
	}

	free(na->node);
}

/*
 * SIDE EFFECT
 *	On successful execution ``nl'' object doesn't contain any keys.
 */
int
nodes_append(struct node_array *na, struct newline *nl)
{
	struct node *p, **oldnode;
	size_t i;

	if (nl->vallen == 0) {
		warnx("key \"%s\" doesn't contain value. Ignored", nl->key[0]);

		/* Remove all keys */
		newline_final(nl);
		newline_init(nl);
		return (0);
	}

	oldnode = na->node;
	na->node = (struct node **)realloc(na->node,
	    sizeof(*na->node) * (na->n_node + nl->n_key));
	if (na->node == NULL) {
		na->node = oldnode;
		warnx("can't allocate memory");
		return (ENOMEM);
	}

	for (i = 0; i < nl->n_key; i++) {
		p = (struct node *)malloc(sizeof(*p));
		if (p == NULL) {
			for (; i > 0; i--)
				free(na->node[i - 1 + na->n_node]);
			warnx("can't allocate memory");
			return (ENOMEM);
		}

		p->key = nl->key[i];
		p->val = nl->val;
		p->vallen = nl->vallen;
		if (i > 0)
			p->orig = na->node[na->n_node];

		na->node[i + na->n_node] = p;
	}

	na->n_node += nl->n_key;
	newline_drain(nl);
	return (0);
}

int
nodes_parse(struct node_array *na, const char *buf, size_t size)
{
	struct newline nl;
	char *p, *lim;
	int error;

	p = (char *)buf;
	lim = p + size;
	while (p < lim) {
		newline_init(&nl);
		error = newline_parse(&nl, p, lim, &p);
		if (error != 0) {
			newline_final(&nl);
			return (error);
		}

		if (nl.n_key > 0) {
			error = nodes_append(na, &nl);
			newline_final(&nl);
			if (error != 0)
				return (error);
		}
	}

	return (0);
}

void
nodes_sort(struct node_array *na)
{
	qsort(na->node, na->n_node, sizeof(*na->node), nodecmp);
}

int
nodes_save(struct node_array *na, FILE *cfile, FILE *hfile)
{
	size_t i, wrsz, index;

	fprintf(hfile, hfile_fmt, inputfile, structname, structname, headername,
	    structname, funcprefix, structname, structname);
	fprintf(cfile, cfile_fmt1, inputfile, headername, structname,
	    structname, structname, structname);

	index = 0;
	for (i = 0; i < na->n_node; i++) {
		if (na->node[i]->orig == NULL) {
			fprintf(cfile, "\t{");
			wrsz = fwrite(na->node[i]->val, na->node[i]->vallen, 1,
			    cfile);
			if (wrsz != 1) {
				warnx("can't write c-file");
				return (EIO);
			}
			fprintf(cfile, "},\n");
			na->node[i]->index = index++;
		}
	}

	fprintf(cfile, cfile_fmt2, structname, structname);

	for (i = 0; i < na->n_node; i++) {
		index = (na->node[i]->orig == NULL) ? na->node[i]->index :
		    na->node[i]->orig->index;
		fprintf(cfile, cfile_fmt2dot5, na->node[i]->key,
		    structname, index);
	}

	fprintf(cfile, cfile_fmt3, structname, funcprefix, structname,
	    structname, funcprefix, structname, structname, structname,
	    (icase) ? "case" : "", structname, structname);
	return (0);
}

static int
validate_code(const char *str)
{
	char *p;

	p = (char *)str;
	while (*p != '\0') {
		if (isalnum(*p) == 0 && *p != '_')
			return (0);
		++p;
	}
	return (1);
}

size_t
read_task_file(const char *filename, char **buf)
{
	struct stat st;
	int fd, error;
	size_t rv, sz, i;
	ssize_t rdsize;

	fd = open(filename, O_RDONLY);
	if (fd == -1) {
		warnx("can't open input file");
		return (0);
	}

	*buf = NULL;
	rv = 0;
	do {
		error = fstat(fd, &st);
		if (error == -1) {
			warnx("can't get size of input file");
			break;
		}
		if (st.st_size < 1 || st.st_size > SIZE_T_MAX) {
			warnx("bad size of input file");
			break;
		}
		sz = st.st_size;

		*buf = (char *)malloc(sz);
		if (*buf == NULL) {
			warnx("can't allocate memory");
			break;
		}

		i = 0;
		do {
			rdsize = read(fd, *buf + i, sz - i);
			if (rdsize < 1) {
				warnx("can't read input file");
				break;
			}
			i += rdsize;
		} while (i < sz);
		rv = sz;
	} while (0);

	if (rv == 0 && *buf != NULL)
		free(*buf);
	close(fd);
	return (rv);
}

int
main(int argc, char **argv)
{
	int ch;
	FILE *cfile, *hfile;
	char *buf;
	size_t inlen;
	struct node_array na;
	char *cname, *hname, *p1, *p2;
	size_t namlen;
	int error;

	while ((ch = getopt(argc, argv, "?chp:")) != -1) {
		switch (ch) {
		case 'c':
			icase = 1;
			break;
		case 'p':
			funcprefix = optarg;
			if (!validate_code(funcprefix))
				errx(1, "invalid prefix");
			break;
		case '?':
		case 'h':
		default:
			usage();
			/* NOTREACHED */
		}
	}

	argc -= optind;
	argv += optind;

	if (argc < 2)
		usage();

	structname = argv[0];
	headername = argv[1];
	inputfile = argv[2];

	if (*structname == '\0' || !validate_code(structname))
		errx(1, "invalid structure name");
	else if (*headername == '\0')
		errx(1, "invalid header file name");

	p1 = strrchr(argv[2], '/');
	if (p1 == NULL)
		p1 = argv[2];
	else
		p1++;
	p2 = strchr(p1, '.');
	if (p2 == NULL)
		p2 = p1 + strlen(p1);

	inlen = read_task_file(argv[2], &buf);
	if (inlen == 0)
		exit(1);

	nodes_init(&na);
	error = nodes_parse(&na, buf, inlen);
	free(buf);
	if (error != 0)
		exit(1);

	if (na.n_node == 0)
		errx(1, "input file doesn't contain key/value");

	nodes_sort(&na);

	namlen = p2 - p1;
	cname = (char *)malloc(namlen + 3);
	if (cname == NULL)
		errx(1, "can't allocate memory");
	hname = (char *)malloc(namlen + 3);
	if (hname == NULL)
		errx(1, "can't allocate memory");
	memcpy(cname, p1, namlen);
	memcpy(hname, p1, namlen);
	memcpy(cname + namlen, ".c", 3);
	memcpy(hname + namlen, ".h", 3);

	cfile = fopen(cname, "w");
	free(cname);
	if (cfile == NULL)
		errx(1, "can't open output c-file");
	hfile = fopen(hname, "w");
	free(hname);
	if (hfile == NULL)
		errx(1, "can't open output h-file");

	error = nodes_save(&na, cfile, hfile);
	nodes_final(&na);
	if (error != 0)
		exit(1);

	fclose(hfile);
	fclose(cfile);

	exit(0);
}
